# ruff: noqa: F401

from . import api, models, types, warnings
from ._load_udf import load
from ._options import _env, options
from ._public_api import ingest, ingest_nongeospatial
from ._run import run
from ._secrets import secrets
from ._udf import udf
from ._utils import utils

# TODO: fused.{download, get_chunk_from_table, get_chunks_metadata} are deprecated here
from .core import cache, download, file_path, get_chunk_from_table, get_chunks_metadata

# Version is set to 0.0.0 as the version string is automatically generated by
# https://github.com/mtkennerly/poetry-dynamic-versioning
__version__ = "1.10.3"



import re
import warnings
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    Literal,
    Optional,
    Union,
    overload,
)

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd
    import shapely
    import xarray as xr

from .core import (
    run_file,
    run_file_async,
    run_shared_file,
    run_shared_file_async,
    run_shared_tile,
    run_shared_tile_async,
    run_tile,
    run_tile_async,
)
from .core._impl._realtime_ops_impl import default_run_engine
from .core._impl._reimports import GeoPandasUdfV2, UdfAccessToken, UdfJobStepConfig

ResultType = Union["xr.Dataset", "pd.DataFrame", "gpd.GeoDataFrame"]


OLD_TOKEN_REGEX = re.compile("^[a-f0-9]{64}$")


def _is_udf_token(maybe_token: str):
    if OLD_TOKEN_REGEX.match(maybe_token):
        return True
    return maybe_token.startswith("UDF_") or maybe_token.startswith("fsh_")


@overload
def run(
    email_or_udf_or_token: Union[
        str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken
    ],
    /,
    udf_name: Optional[str],
    *,
    udf: Optional[GeoPandasUdfV2],
    job_step: Optional[UdfJobStepConfig],
    token: Optional[str],
    udf_email: Optional[str],
    x: Optional[int],
    y: Optional[int],
    z: Optional[int],
    _lat: Optional[float],
    _lng: Optional[float],
    bbox: Union["gpd.GeoDataFrame", "shapely.Geometry", None],
    sync: bool = False,
    engine: Optional[Literal["realtime", "batch", "local"]],
    type: Optional[Literal["tile", "file"]],
    parameters: Optional[Dict[str, Any]] = None,
    **kw_parameters,
) -> Coroutine[ResultType, None, None]:
    ...


@overload
def run(
    email_or_udf_or_token: Union[str, None, UdfJobStepConfig, GeoPandasUdfV2],
    /,
    udf_name: Optional[str],
    *,
    udf: Optional[GeoPandasUdfV2],
    job_step: Optional[UdfJobStepConfig],
    token: Optional[str],
    udf_email: Optional[str],
    x: Optional[int],
    y: Optional[int],
    z: Optional[int],
    sync: bool,
    engine: Optional[Literal["realtime", "batch", "local"]],
    type: Optional[Literal["tile", "file"]],
    parameters: Optional[Dict[str, Any]] = None,
    **kw_parameters,
) -> ResultType:
    ...


def run(
    email_or_udf_or_token: Union[
        str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken
    ] = None,
    /,
    udf_name: Optional[str] = None,
    *,
    udf: Optional[GeoPandasUdfV2] = None,
    job_step: Optional[UdfJobStepConfig] = None,
    token: Optional[str] = None,
    udf_email: Optional[str] = None,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: bool = True,
    engine: Optional[Literal["realtime", "batch", "local"]] = None,
    type: Optional[Literal["tile", "file"]] = None,
    parameters: Optional[Dict[str, Any]] = None,
    **kw_parameters,
):
    """
    Executes a user-defined function (UDF) with various execution and input options.

    This function supports executing UDFs in different environments (realtime, batch, local),
    with different types of inputs (tile coordinates, geographical bounding boxes, etc.), and
    allows for both synchronous and asynchronous execution. It dynamically determines the execution
    path based on the provided parameters.

    Args:
        email_or_udf_or_token: A string that can either be a UDF name, a UDF token, or a direct
            reference to a UDF object. It can also be a UdfJobStepConfig object for detailed
            configuration, or None to specify UDF details in other parameters.
        udf_name: The name of the UDF to execute.
        udf: A GeoPandasUdfV2 object for direct execution.
        job_step: A UdfJobStepConfig object for detailed execution configuration.
        token: A token representing a shared UDF.
        udf_email: The email associated with the UDF.
        x, y, z: Tile coordinates for tile-based UDF execution.
        sync: If True, execute the UDF synchronously. If False, execute asynchronously.
        engine: The execution engine to use ('realtime', 'batch', or 'local').
        type: The type of UDF execution ('tile' or 'file').
        parameters: Additional parameters to pass to the UDF.
        **kw_parameters: Additional parameters to pass to the UDF.

    Raises:
        ValueError: If the UDF is not specified or is specified in more than one way.
        TypeError: If the first parameter is not of an expected type.
        Warning: Various warnings are issued for ignored parameters based on the execution path chosen.

    Returns:
        The result of the UDF execution, which varies based on the UDF and execution path.

    Examples:


        Run a UDF saved in the Fused system:
        ```py
        fused.run(udf_email="username@fused.io", udf_name="my_udf_name")
        ```

        Run a UDF saved in GitHub:
        ```py
        loaded_udf = fused.load("https://github.com/fusedio/udfs/tree/main/public/Building_Tile_Example")
        fused.run(udf=loaded_udf, bbox=bbox)
        ```

        Run a UDF saved in a local directory:
        ```py
        loaded_udf = fused.load("/Users/local/dir/Building_Tile_Example")
        fused.run(udf=loaded_udf, bbox=bbox)
        ```

    Note:
        This function dynamically determines the execution path and parameters based on the inputs.
        It is designed to be flexible and support various UDF execution scenarios.
    """
    from fused._optional_deps import HAS_GEOPANDAS, HAS_MERCANTILE, HAS_SHAPELY

    if email_or_udf_or_token is not None:
        if udf is not None:
            warnings.warn(
                "udf parameter is being ignored in favor of the first positional parameter.",
            )
        if job_step is not None:
            warnings.warn(
                "job_step parameter is being ignored in favor of the first positional parameter.",
            )
        if token is not None:
            warnings.warn(
                "token parameter is being ignored in favor of the first positional parameter.",
            )
        if udf_email is not None:
            warnings.warn(
                "udf_email parameter is being ignored in favor of the first positional parameter.",
            )

        if isinstance(email_or_udf_or_token, UdfJobStepConfig):
            job_step = email_or_udf_or_token
        elif isinstance(email_or_udf_or_token, GeoPandasUdfV2):
            udf = email_or_udf_or_token
        elif isinstance(email_or_udf_or_token, UdfAccessToken):
            token = email_or_udf_or_token.token
        elif isinstance(email_or_udf_or_token, str):
            if "/" in email_or_udf_or_token:
                udf_email, udf_name = email_or_udf_or_token.split("/", maxsplit=1)
            elif "@" in email_or_udf_or_token:
                udf_email = email_or_udf_or_token
            elif _is_udf_token(email_or_udf_or_token):
                token = email_or_udf_or_token
            else:
                # This will actually be the udf name, not the user's email
                udf_email = email_or_udf_or_token
                udf_name = None
        else:
            raise TypeError(
                "Could not detect UDF from first parameter. It should be a string, UdfJobStepConfig, or BaseUdf object."
            )

    ways_of_specifying_udfs = sum([bool(x) for x in [udf, job_step, token, udf_email]])
    if ways_of_specifying_udfs == 0:
        raise ValueError("No UDF specified")
    if ways_of_specifying_udfs != 1:
        raise ValueError("UDF was specified in more than one way, which is invalid")

    local_tile_bbox: Optional["gpd.GeoDataFrame"] = None
    xyz_ignored = False
    if (
        x is not None
        and y is not None
        and z is not None
        and HAS_MERCANTILE
        and HAS_GEOPANDAS
        and HAS_SHAPELY
    ):
        import geopandas as gpd
        import mercantile
        import shapely

        tile_bounds = mercantile.bounds(x, y, z)
        local_tile_bbox = gpd.GeoDataFrame(
            {"x": [x], "y": [y], "z": [z]},
            geometry=[shapely.box(*tile_bounds)],
            crs=4326,
        )
    elif x is not None or y is not None or z is not None:
        xyz_ignored = True

    if x is not None and y is not None and z is not None:
        if type is None:
            type = "tile"
        elif type != "tile":
            warnings.warn(
                "x, y, z specified but UDF type is not 'tile', so they will be ignored",
            )
    else:
        if type is None:
            type = "file"
        elif type != "file":
            raise ValueError(
                "x, y, z not specified but type is 'tile', which is an invalid configuration. You must specify x, y, and z."
            )

    if udf is not None:
        job_step = UdfJobStepConfig(udf=udf)

    parameters = {
        **kw_parameters,
        **(parameters if parameters is not None else {}),
    }

    dispatch: dict[
        tuple[
            Literal["sync", "async"],
            Literal["tile", "file"],
            Literal["saved", "token", "local_job_step"],
            Literal["realtime", "local", "batch"],
        ],
        Optional[Callable],
    ] = {
        # Saved UDF
        ("sync", "tile", "saved", "realtime"): partial(
            run_tile, udf_email, udf_name, x=x, y=y, z=z, **parameters
        ),
        ("async", "tile", "saved", "realtime"): partial(
            run_tile_async, udf_email, udf_name, x=x, y=y, z=z, **parameters
        ),
        ("sync", "file", "saved", "realtime"): partial(
            run_file, udf_email, udf_name, **parameters
        ),
        ("async", "file", "saved", "realtime"): partial(
            run_file_async, udf_email, udf_name, **parameters
        ),
        # shared UDF token
        ("sync", "tile", "token", "realtime"): partial(
            run_shared_tile, token, x=x, y=y, z=z, **parameters
        ),
        ("async", "tile", "token", "realtime"): partial(
            run_shared_tile_async, token, x=x, y=y, z=z, **parameters
        ),
        ("sync", "file", "token", "realtime"): partial(
            run_shared_file, token, **parameters
        ),
        ("async", "file", "token", "realtime"): partial(
            run_shared_file_async, token, **parameters
        ),
        # Local job step, which includes locally held code
        ("sync", "tile", "local_job_step", "realtime"): lambda: job_step.run_tile(
            x=x, y=y, z=z, **parameters
        ),
        (
            "async",
            "tile",
            "local_job_step",
            "realtime",
        ): lambda: job_step.run_tile_async(x=x, y=y, z=z, **parameters),
        ("sync", "file", "local_job_step", "realtime"): lambda: job_step.run_file(
            **parameters
        ),
        (
            "async",
            "file",
            "local_job_step",
            "realtime",
        ): lambda: job_step.run_file_async(**parameters),
        ("sync", "tile", "local_job_step", "local"): lambda: job_step.run_local(
            local_tile_bbox, **parameters
        ),
        ("async", "tile", "local_job_step", "local"): lambda: job_step.run_local(
            local_tile_bbox, **parameters
        ),
        ("sync", "file", "local_job_step", "local"): lambda: job_step.run_local(
            **parameters
        ),
        ("async", "file", "local_job_step", "local"): lambda: job_step.run_local(
            **parameters
        ),
        ("sync", "tile", "local_job_step", "batch"): None,
        ("sync", "file", "local_job_step", "batch"): lambda: job_step.set_udf(
            job_step.udf, parameters=parameters
        ).run_remote(),
    }

    udf_storage = (
        "saved"
        if udf_email is not None
        else (
            "token"
            if token is not None
            else ("local_job_step" if job_step is not None else None)
        )
    )
    if udf_storage is None:
        raise ValueError("No UDF specified")

    if engine is None:
        if udf_storage in ["token", "saved"]:
            engine = "realtime"
        else:
            engine = default_run_engine()

    if xyz_ignored:
        if x is None or y is None or z is None:
            warnings.warn(
                "x, y, z arguments will be ignored because one of them is None"
            )
        elif engine == "local":
            # This warning doesn't matter on realtime because we will just put the x/y/z into the URL
            warnings.warn(
                "x, y, z arguments will be ignored because the following packages were not all found: mercantile shapely geopandas"
            )

    dispatch_params = ("sync" if sync else "async", type, udf_storage, engine)

    # Ellipsis is the sentinal value for not found in the dictionary at all
    fn = dispatch.get(dispatch_params, ...)

    if fn is Ellipsis:
        if udf_storage == "token" and engine != "realtime":
            raise ValueError("UDF tokens can only be called on the realtime engine.")
        elif udf_storage == "saved" and engine != "realtime":
            raise ValueError(
                "Saved UDFs can only be called on the realtime engine. To use another engine, load the UDF locally first."
            )
        else:
            raise ValueError(
                f"Could not determine how to call with settings: {dispatch_params}"
            )
    if fn is None:
        raise ValueError(f"Call type is not yet implemented: {dispatch_params}")

    return fn()



from __future__ import annotations

import ast
import inspect
import sys
import time
import warnings
from contextlib import ExitStack
from textwrap import dedent, indent
from typing import Any, Callable, Dict, Optional

import pandas as pd

import fused
from fused._udf.context import ExecutionContextProtocol, context
from fused._udf.state import decorator_src_override_context, noop_decorators_context
from fused.core._impl._exec_globals_impl import make_exec_globals
from fused.models.schema import Schema
from fused.models.udf import EMPTY_UDF, AnyBaseUdf
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.models.udf.output import Output, PandasOutput
from fused.warnings import FusedWarning


def _add_special_kwargs(
    fn: Callable,
    args: Dict[str, Any],
    output: Output,
    context: ExecutionContextProtocol,
):
    signature = inspect.signature(fn)
    param_names = signature.parameters.keys()
    # note: we don't check that this is a kwarg on the function, so if it's the first parameter it could cause
    # a call failure.
    if "output" in param_names:
        args["output"] = output
    if "context" in param_names:
        args["context"] = context

    return args


def execute_against_sample(
    udf: AnyBaseUdf,
    input: Any,
    update_schema: bool = True,
    validate_output: bool = True,
    validate_imports: Optional[bool] = None,
    **kwargs,
) -> UdfEvaluationResult:
    if udf is EMPTY_UDF:
        raise ValueError("Empty UDF cannot be evaluated. Use `set_udf` to set a UDF.")

    output = Output()
    try:
        # Define custom function in environment

        # Wrap cell in a function
        # This is believed to fix scoping issues with closures and list comprehensions
        # See also https://stackoverflow.com/a/2749806,
        # https://stackoverflow.com/q/38318370
        # {cell_text} is not itself indented because we need to make sure that every
        # line of {cell_text} is indented, not just the first.
        FN_RESULT_NAME = "__fused_fn_result"
        wrapped_cell = """\
            def __fused_exec():
            {cell_text}

                return {entrypoint}

            {FN_RESULT_NAME} = __fused_exec()
            """
        original_src = udf.code
        src = dedent(wrapped_cell).format(
            cell_text=indent(dedent(original_src), " " * 4),
            entrypoint=udf.entrypoint,
            FN_RESULT_NAME=FN_RESULT_NAME,
        )

        exec_locals = {}
        exec_globals = make_exec_globals()

        # Store names of modules prior to rendering Headers
        preceding_sys_modules = list(sys.modules.keys())

        # Validate import stamements correspond to valid modules
        validate_imports_whitelist(udf, validate_imports=validate_imports)

        with ExitStack() as stack:
            stack.enter_context(noop_decorators_context(True))

            # Add headers to sys.meta_path
            if udf.headers is not None:
                for header in udf.headers:
                    stack.enter_context(
                        header._register_custom_finder(preceding_sys_modules)
                    )

            exec(src, exec_globals, exec_locals)

            _fn = exec_locals[FN_RESULT_NAME]

            kwargs_to_pass = {**udf.parameters, **kwargs}
            _add_special_kwargs(
                _fn, args=kwargs_to_pass, output=output, context=context
            )
            args_to_pass = (
                input.as_udf_args() if hasattr(input, "as_udf_args") else input
            )

            # Run UDF
            time_start = time.perf_counter()
            _output = _fn(*args_to_pass, **kwargs_to_pass)
            time_end = time.perf_counter()
            time_taken_seconds = time_end - time_start

            if not validate_output:
                return _output

            if _output is not None:
                output.data = _output

    except Exception:
        raise

    new_output = _transform_output(output=output)

    if new_output is None:
        # TODO: Assumes table_schema is present, which doesn't match type
        if udf.table_schema not in [None, {}, Schema()]:
            warnings.warn(
                "UDF is configured with a schema but returns `None`. An empty schema was set for this execution.",
                FusedWarning,
            )

        return UdfEvaluationResult(
            data=None,
            sidecar=None,
            udf=udf,
            table_schema={},
            time_taken_seconds=time_taken_seconds,
        )

    # Validate the dataframe after assigning it to the output variable in user_ns
    # so the user can inspect the output if anything is wrong.
    new_output.validate_data_with_schema()

    if update_schema and udf.table_schema is None:
        udf.table_schema = new_output.table_schema

    return UdfEvaluationResult(
        data=new_output.data,
        sidecar=new_output.sidecar_output,
        udf=udf,
        table_schema=new_output.table_schema,
        time_taken_seconds=time_taken_seconds,
    )


def execute_for_decorator(udf: AnyBaseUdf) -> AnyBaseUdf:
    """Evaluate a UDF for the purpose of getting the UDF object out of it."""
    # Define custom function in environment

    # This is a stripped-down version of execute_against_sample, above.

    FN_RESULT_NAME = "__fused_fn_result"
    wrapped_cell = """\
        def __fused_exec():
        {cell_text}

            return {entrypoint}

        {FN_RESULT_NAME} = __fused_exec()
        """
    original_src = udf.code
    src = dedent(wrapped_cell).format(
        cell_text=indent(dedent(original_src), " " * 4),
        entrypoint=udf.entrypoint,
        FN_RESULT_NAME=FN_RESULT_NAME,
    )

    exec_locals = {}
    exec_globals = make_exec_globals()

    with ExitStack() as stack:
        stack.enter_context(decorator_src_override_context(original_src))

        # Add headers to sys.meta_path
        if udf.headers is not None:
            for header in udf.headers:
                stack.enter_context(header._register_custom_finder())

        exec(src, exec_globals, exec_locals)

        _fn = exec_locals[FN_RESULT_NAME]

        return _fn


def _transform_output(output: Output) -> PandasOutput:
    # TODO: Support PyArrow
    if isinstance(output, PandasOutput):
        return output

    if isinstance(output.data, pd.DataFrame):
        # Force all column names to be strings
        output.data.columns = [str(x) for x in output.data.columns]

        # Note that we don't pass a _new_ schema argument here
        # - If the user defined a schema onto output.table_schema, that will be
        #   preserved.
        # - If the user did not define output.table_schema, then output.table_schema
        #   will be None, and PandasOutput will default to a schema inferred from
        #   the DataFrame.
        return PandasOutput(
            data=output.data,
            table_schema=output.table_schema,
            sidecar_output=output.sidecar_output,
            skip_fused_index_validation=output.skip_fused_index_validation,
        )
    elif output.data is None:
        return None

    raise TypeError(f"Unexpected result type {type(output.data)}")


def validate_imports_whitelist(
    udf: AnyBaseUdf, validate_imports: Optional[bool] = None
):
    # Skip import validation if the option is set
    if not fused.options.default_validate_imports and validate_imports is not True:
        return

    # Skip import validation if not logged in
    if not fused.api.AUTHORIZATION.is_configured():
        return

    from fused._global_api import get_api

    # Get the dependency whitelist from the cached API endpoint
    api = get_api()
    package_dependencies = api.dependency_whitelist()

    # Initialize a list to store the import statements
    import_statements = []

    # Parse the source code into an AST
    tree = ast.parse(udf.code)

    # Traverse the AST to find import statements
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                import_statements.append(alias.name)
        elif isinstance(node, ast.ImportFrom):
            module_name = node.module
            import_statements.append(module_name)

    # Check for unavailable modules
    header_modules = [header.module_name for header in udf.headers]
    fused_modules = ["fused"]  # assume fused is always available
    available_modules = (
        list(package_dependencies["dependency_whitelist"].keys())
        + header_modules
        + fused_modules
    )
    unavailable_modules = []
    for import_statement in import_statements:
        if import_statement.split(".", 1)[0] not in available_modules:
            unavailable_modules.append(import_statement)

    if unavailable_modules:
        raise ValueError(
            f"The following imports in the UDF might not be available: {repr(unavailable_modules)}. Please check the UDF headers and imports and try again."
        )

    # TODO: check major versions for some packages



from io import BytesIO
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd


def parquet_to_df(b: bytes) -> Union["pd.DataFrame", "gpd.GeoDataFrame"]:
    try:
        import geopandas as gpd

        return gpd.read_parquet(BytesIO(b))
    except:  # noqa: E722
        import pandas as pd

        return pd.read_parquet(BytesIO(b))



from __future__ import annotations

import base64
import hashlib
import json
import os
import secrets
import webbrowser
from datetime import datetime, timedelta, timezone
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from textwrap import dedent
from threading import RLock
from typing import Dict, Optional
from urllib.parse import parse_qs, urlencode, urlparse

import requests
from pydantic import BaseModel, ValidationError

from fused._environment import is_pyodide
from fused._options import options as OPTIONS

# The authorize URL is used for the initial login flow. This is intended to be opened in
# the user's web browser for them to sign in.
AUTHORIZE_URL = "https://dev-tjcykxcetrz6bps6.us.auth0.com/authorize"

# The token url is used for programmatic access to generate access and refresh tokens
OAUTH_TOKEN_URL = "https://dev-tjcykxcetrz6bps6.us.auth0.com/oauth/token"

LOGOUT_URL = "https://dev-tjcykxcetrz6bps6.us.auth0.com/oidc/logout"

# The client id, client secret, and audience identifies a specific application and API
CLIENT_ID = "CXiwKZQmmyo0rqXZY7pzBgfsF7AL2A9l"
CLIENT_SECRET = "FVNz012KgNmqITYnCCOM8Q1Nt81W_DO4SeCRgVsftREKTWpzZU522nia5TdSNv8h"
AUDIENCE = "fused-python-api"

# This redirect uri is passed to the authorize URL as a url parameter. This localhost
# uri is used to intercept the "code" generated from the authorization
LOCAL_REDIRECT_URI = "http://localhost:3000"

# The offline_access scope is necessary to be able to fetch refresh tokens
# The other scopes are useful to access identifying information in the retrieved JWT
SCOPES = ["openid", "email", "name", "offline_access"]

# The path where the refresh token is saved on disk
CREDENTIALS_PATH = Path("~/.fused/credentials").expanduser()


_refresh_token_lock = RLock()


class BadCredentialsException(Exception):
    def __init__(self, message=""):
        self.http_status_code = 401
        self.detail = "Bad credentials"
        if message:
            self.detail = f"{self.detail}: {message}"


class Credentials(BaseModel):
    """A dataclass representation of OAuth2 credentials"""

    access_token: str
    auth_scheme: str
    refresh_token: str
    id_token: str
    scope: str
    expires_in: int
    expires_at: datetime

    @classmethod
    def authenticate(cls) -> Credentials:
        token_dict = authenticate()
        credentials = cls(**token_dict)
        credentials.save_to_disk()
        return credentials

    @classmethod
    def from_disk(cls) -> Credentials:
        if CREDENTIALS_PATH.exists():
            with open(CREDENTIALS_PATH, "r") as file:
                content = file.read()
            try:
                return cls.model_validate_json(content)
            except ValidationError as e:
                raise ValueError(f"Invalid credentials file: {e}")
        raise ValueError("Credentials file does not exist")

    @classmethod
    def from_disk_or_authenticate(cls) -> Credentials:
        try:
            return cls.from_disk()
        except ValueError:
            return cls.authenticate()

    def save_to_disk(self) -> None:
        if OPTIONS.save_user_settings:
            CREDENTIALS_PATH.parent.mkdir(parents=True, exist_ok=True)
            with open(CREDENTIALS_PATH, "w") as f:
                f.write(self.model_dump_json())

    @classmethod
    def from_token_response(cls, token_dict: Dict) -> Credentials:
        # Add an expires_at field in the dictionary for next time it's loaded
        token_dict["expires_at"] = (
            datetime.now(timezone.utc) + timedelta(seconds=token_dict["expires_in"] - 1)
        ).isoformat()
        token_dict["auth_scheme"] = "Bearer"

        credentials = cls(**token_dict)
        credentials.save_to_disk()

        return credentials

    def refresh_if_needed(self) -> Credentials:
        # Do not call the refresh token from multiple threads at once. If this happens,
        # the rate limit can accidentally be exceeded.
        # This could be optimized with double-checked locking, but the Wikipedia page
        # points out there could be an issue with the object being in a partially updated
        # state.
        # https://en.wikipedia.org/wiki/Double-checked_locking
        with _refresh_token_lock:
            if self.expires_at < datetime.now(timezone.utc):
                # Credentials are expired; we need to refresh
                token_dict = refresh_token(self.refresh_token)
                new_credentials = Credentials(**token_dict)
                self.__dict__.update(new_credentials.__dict__)

        return self


class BearerAccessToken:
    def __init__(self, access_token: str, auth_scheme: str = "Bearer"):
        self.access_token = access_token
        self.auth_scheme = auth_scheme


class MaybeInitializedAuthorization:
    """OAuth2 credentials or Bearer access token obtained from the global
    context that may or may not have been initialized.
    """

    _credentials: Credentials | None
    _bearer_access_token: BearerAccessToken | None

    def __init__(self) -> None:
        self._credentials = None
        if (token := self._maybe_retrieve_auth_token_from_global_context()) is not None:
            self._bearer_access_token = token
            return

        try:
            self._credentials = Credentials.from_disk()
        except ValueError:
            if OPTIONS.prompt_to_login:
                auth_msg = """\
                    Credentials not found on disk. Authenticate with:

                    from fused import NotebookCredentials
                    credentials = NotebookCredentials()
                """
                print(dedent(auth_msg))

    def initialize(self) -> None:
        """Force initialization of credentials."""
        _ = self.credentials

    def _maybe_retrieve_auth_token_from_global_context(
        self,
    ) -> BearerAccessToken | None:
        from fused.core._impl._context_impl import (
            context_get_auth_scheme_and_token,
            context_in_realtime,
        )

        if not context_in_realtime():
            return

        # SECURITY: We are currently inside RT2 and access to RT2 is already
        # authenticated via fused-server. But RT2 may send requests back to
        # fused-server for invocations of fused.run within a UDF. And for
        # fused-server to be able to authorize the request correctly, we need to
        # ensure that RT2 is including the access token back in the request.
        #
        # As a result, directly injecting the token back without additional
        # validation is acceptable here.
        maybe_auth_from_context = context_get_auth_scheme_and_token()
        if (
            maybe_auth_from_context
            and maybe_auth_from_context[0]
            and maybe_auth_from_context[1]
        ):
            return BearerAccessToken(
                maybe_auth_from_context[1], maybe_auth_from_context[0]
            )

        # Fallback to using token from env if the global context does not have
        # the token. But also ensure we only accept tokens of type Fused-Rtenv-Bearer.
        auth_scheme = os.getenv("FUSED_AUTH_TOKEN_TYPE", "")
        token = os.getenv("FUSED_AUTH_TOKEN", "")
        if auth_scheme == "Fused-Rtenv-Bearer" and token != "":
            return BearerAccessToken(token, auth_scheme)

        raise BadCredentialsException("AuthN configuration error.")

    @property
    def credentials(self) -> Credentials | BearerAccessToken:
        """
        Retrieve valid credentials, initializing them from global context or
        authenticating from scratch if needed.
        """
        if (token := self._maybe_retrieve_auth_token_from_global_context()) is not None:
            self._bearer_access_token = token
            return self._bearer_access_token

        existing_credentials = (
            self._credentials or Credentials.from_disk_or_authenticate()
        )

        existing_credentials = existing_credentials.refresh_if_needed()
        self._credentials = existing_credentials
        return existing_credentials

    def set_credentials(self, credentials: Credentials) -> None:
        self._credentials = credentials

    def is_configured(self) -> bool:
        try:
            if OPTIONS.no_login:
                return False

            if self._credentials is not None:
                return True

            return Credentials.from_disk() is not None
        except Exception:
            return False


AUTHORIZATION: MaybeInitializedAuthorization = MaybeInitializedAuthorization()
"""Global credentials."""


# TODO: remove some of these functions below
def authenticate():
    if is_pyodide():
        # We cannot use this method to log in in Pyodide, so show an error message instead.
        raise PermissionError(
            "Fused credentials not found -- please log in.\n\nIf you are already logged in, please make sure that `fused` is installed through the App's requirements."
        )

    print(
        'Existing credentials not found on disk: Please check your browser to log in. If using a notebook, use "credentials = fused.api.NotebookCredentials()" to log in.'
    )
    code_verifier = secrets.token_urlsafe(48)
    code_challenge = get_code_challenge(code_verifier)

    params = {
        "audience": AUDIENCE,
        "scope": " ".join(SCOPES),
        "response_type": "code",
        "client_id": CLIENT_ID,
        "redirect_uri": LOCAL_REDIRECT_URI,
        "code_challenge_method": "S256",
        "code_challenge": code_challenge,
    }
    url = f"{AUTHORIZE_URL}?{urlencode(params)}"
    code = handle_redirect(url)

    token_data = {
        "client_id": CLIENT_ID,
        "grant_type": "authorization_code",
        "audience": AUDIENCE,
        "client_secret": CLIENT_SECRET,
        "code": code,
        "redirect_uri": LOCAL_REDIRECT_URI,
        "code_verifier": code_verifier,
    }
    token_response = requests.post(
        OAUTH_TOKEN_URL, json=token_data, timeout=OPTIONS.request_timeout
    )
    token_response.raise_for_status()
    token_dict = token_response.json()

    # Add an expires_at field in the dictionary for next time it's loaded
    token_dict["expires_at"] = (
        datetime.now(timezone.utc) + timedelta(seconds=token_dict["expires_in"] - 1)
    ).isoformat()
    token_dict["auth_scheme"] = "Bearer"

    save_token_to_disk(token_dict)
    return token_dict


def logout():
    """Open the user's browser to the Auth0 logout page."""
    webbrowser.open(LOGOUT_URL)


def save_token_to_disk(token: dict):
    if OPTIONS.save_user_settings:
        CREDENTIALS_PATH.parent.mkdir(parents=True, exist_ok=True)
        with open(CREDENTIALS_PATH, "w") as f:
            json.dump(token, f)


def delete_token_from_disk():
    CREDENTIALS_PATH.unlink(missing_ok=True)


def get_code_challenge(code_verifier: str) -> str:
    """Take an input string and hash it to generate a challenge string

    Refer to https://auth0.com/docs/get-started/authentication-and-authorization-flow/call-your-api-using-the-authorization-code-flow-with-pkce
    """
    code_challenge_digest = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    code_challenge = base64.urlsafe_b64encode(code_challenge_digest).decode("utf-8")
    return code_challenge.replace("=", "")


def handle_redirect(authorize_url: str) -> str:
    """Open the authorization url and intercept its redirect

    The redirection from the `/authorize` endpoint includes a code that can be used
    against the `/oauth/token` endpoint to fetch a refresh and access token.
    """
    code: Optional[str] = None

    # TODO: this request handler should also support when a user declines the auth flow
    class RequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            nonlocal code

            query_string = urlparse(self.path).query
            parsed_qs = parse_qs(query_string)

            assert len(parsed_qs["code"]) == 1
            code = parsed_qs["code"][0]

            self.send_response(200)
            self.send_header("Content-Type", "text/html")
            self.end_headers()

            text = "Success! You can now close this tab and return to Python."
            self.wfile.write(text.encode())

        def log_message(self, format, *args):
            # This is overwritten to prevent the server from logging requests to the
            # console
            return

    # TODO: support other ports if 3000 is already taken?
    server = HTTPServer(("localhost", 3000), RequestHandler)
    webbrowser.open(authorize_url)

    # Note that this only handles _one_ request, but that's all it should need
    server.handle_request()

    assert code is not None
    return code


def refresh_token(refresh_token: str):
    """Generate a new access_token using a refresh token"""
    token_data = {
        "client_id": CLIENT_ID,
        "grant_type": "refresh_token",
        "client_secret": CLIENT_SECRET,
        # TODO: Is this needed? This won't be right for notebook
        "redirect_uri": LOCAL_REDIRECT_URI,
        "refresh_token": refresh_token,
    }
    token_response = requests.post(
        OAUTH_TOKEN_URL, json=token_data, timeout=OPTIONS.request_timeout
    )
    token_response.raise_for_status()
    token_dict = token_response.json()

    # Add an expires_at field in the dictionary for next time it's loaded
    token_dict["expires_at"] = (
        datetime.now(timezone.utc) + timedelta(seconds=token_dict["expires_in"] - 1)
    ).isoformat()
    token_dict["refresh_token"] = refresh_token
    token_dict["auth_scheme"] = "Bearer"

    save_token_to_disk(token_dict)
    return token_dict



from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    import geopandas as gpd
    import mercantile
    import shapely

TileXYZ = TypeVar("TileXYZ", bound="mercantile.Tile")
TileGDF = TypeVar("TileGDF", bound="gpd.GeoDataFrame")
Bbox = TypeVar("Bbox", bound="shapely.geometry.polygon.Polygon")



import warnings
from pathlib import Path
from typing import Any, BinaryIO, Iterable, Literal, Optional, Union, overload

from fused.models.api.job import UdfJobStepConfig
from fused.warnings import FusedUdfWarning

from .base_udf import UdfType
from .udf import GeoPandasUdfV2


class GeoPandasUdfV2Callable(GeoPandasUdfV2):
    type: Literal[UdfType.GEOPANDAS_V2] = UdfType.GEOPANDAS_V2
    """This class is returned from `@fused.udf` and represents
    a UDF that can be instantiated into a job."""

    def to_file(self, where: Union[str, Path, BinaryIO], *, overwrite: bool = False):
        """Write the UDF to disk or the specified file-like object.

        The UDF will be written as a Zip file.

        Args:
            where: A path to a file or a file-like object.

        Keyword Args:
            overwrite: If true, overwriting is allowed.
        """
        job = self()
        job.export(where, how="zip", overwrite=overwrite)

    def to_directory(self, where: Union[str, Path], *, overwrite: bool = False):
        """Write the UDF to disk as a directory (folder).

        Args:
            where: A path to a directory.

        Keyword Args:
            overwrite: If true, overwriting is allowed.
        """
        job = self()
        job.export(where, how="local", overwrite=overwrite)

    # This is a subclass of GeoPandasUdfV2 so that the job classes can reference
    # GeoPandasUdfV2 without issues. This class is then installed over the
    # GeoPandasUdfV2 type code so that loaded objects get the __call__ methods.

    # List of data input is passed - run that
    @overload
    def __call__(self, *, arg_list: Iterable[Any], **kwargs) -> UdfJobStepConfig:
        ...

    # Nothing is passed - run the UDF once
    @overload
    def __call__(self, *, arg_list: None = None, **kwargs) -> UdfJobStepConfig:
        ...

    def __call__(
        self, *, arg_list: Optional[Iterable[Any]] = None, **kwargs
    ) -> Union[UdfJobStepConfig,]:
        """Create a job from this UDF.

        Args:
            arg_list: A list of records to pass in to the UDF as input.
        """

        with_params = self.model_copy()
        # TODO: Consider using with_parameters here, and validating that "context" and other reserved parameter names are not being passed.
        new_parameters = {**kwargs}
        if new_parameters:
            with_params.parameters = new_parameters

        if arg_list is not None and not len(arg_list):
            warnings.warn(
                "An empty `arg_list` was passed in, no calls to the UDF will be made.",
                FusedUdfWarning,
            )

        return UdfJobStepConfig(
            udf=with_params,
            input=arg_list,
        )



# TODO: This file is no longer the most recent -- use fused.core.run_* instead
# This file is only for running non-saved (code included) UDFs

import json
import time
import warnings
from io import BytesIO
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Optional, Union

import geopandas as gpd
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import requests
from loguru import logger

from fused._environment import infer_display_method
from fused._options import options as OPTIONS
from fused._udf.execute_v2 import _transform_output
from fused.api.api import FusedAPI, resolve_udf_server_url
from fused.core._realtime_ops import serialize_realtime_params
from fused.models import AnyJobStepConfig
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.models.udf.output import Output
from fused.models.udf.udf import AnyBaseUdf
from fused.warnings import FusedWarning

from ..core._impl._realtime_ops_impl import get_recursion_factor


def run_tile(
    x: float,
    y: float,
    z: float,
    data: Optional[
        Union[pd.DataFrame, gpd.GeoDataFrame, pa.Table, str, Path, Any]
    ] = None,
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    client_id: Optional[str] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> UdfEvaluationResult:  # TODO: return png
    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    url = f"{udf_server_url}/api/v1/run/udf/tiles/{z}/{x}/{y}"

    # Headers
    recursion_factor = get_recursion_factor()
    headers = api._generate_headers({"Content-Type": "application/json"})
    headers["Fused-Recursion"] = f"{recursion_factor}"

    # Payload
    post_attr_json = {
        "data_left": data,
        "data_right": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Params
    req_params = {}

    # Make request
    start = time.time()

    r = requests.post(
        url=url,
        params=req_params,
        json=post_attr_json,
        headers=headers,
        timeout=OPTIONS.request_timeout,
    )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


def run(
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    read_options: Optional[Dict] = None,
    client_id: Optional[str] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> pd.DataFrame:
    """Run a UDF.

    Args:
        step_config: AnyJobStepConfig.
        params: Additional parameters to pass to the UDF. Must be JSON serializable.

    Keyword Args:
        print_time: If True, print the amount of time taken in the request.
        read_options: If not None, options for reading `df` that will be passed to GeoPandas.
    """
    # TODO: This function is too complicated

    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    # Note: Custom UDF uses the json POST attribute.
    url = f"{udf_server_url}/api/v1/run/udf"

    # This is the body for when step_config_with_params.type == "udf".
    body = {
        "data_left": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    method = "POST"
    post_attr_json = body

    recursion_factor = get_recursion_factor()
    post_attr_headers = api._generate_headers({"Content-Type": "application/json"})
    post_attr_headers["Fused-Recursion"] = f"{recursion_factor}"

    req_params = {}

    # Make request
    start = time.time()

    r = requests.request(
        method=method,
        url=url,
        params=req_params,
        json=post_attr_json,
        headers=post_attr_headers,
        timeout=OPTIONS.request_timeout,
    )
    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


async def run_tile_async(
    x: float,
    y: float,
    z: float,
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    client_id: Optional[str] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> UdfEvaluationResult:  # TODO: return png
    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    url = f"{udf_server_url}/api/v1/run/udf/tiles/{z}/{x}/{y}"

    # Headers
    recursion_factor = get_recursion_factor()
    headers = api._generate_headers({"Content-Type": "application/json"})
    headers["Fused-Recursion"] = f"{recursion_factor}"

    # Payload
    post_attr_json = {
        "data_left": None,
        "data_right": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Params
    req_params = {}

    # Make request
    start = time.time()

    if OPTIONS.pyodide_async_requests:
        import pyodide.http
        import yarl

        url_with_params = yarl.URL(url, encoded=True).with_query(req_params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            method="POST",
            headers=headers,
            body=json.dumps(post_attr_json),
            # TODO: timeout
        )
    else:
        import aiohttp

        async with aiohttp.ClientSession() as session:
            r = await session.post(
                url=url,
                params=req_params,
                json=post_attr_json,
                headers=headers,
                # TODO: timeout
            )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return await _process_response_async(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


async def run_async(
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    client_id: Optional[str] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> pd.DataFrame:
    """Run a UDF over a DataFrame.

    Args:
        step_config: AnyJobStepConfig.
        params: Additional parameters to pass to the UDF. Must be JSON serializable.

    Keyword Args:
        print_time: If True, print the amount of time taken in the request.
    """
    # TODO: This function is too complicated

    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    # Note: Custom UDF uses the json POST attribute.
    url = f"{udf_server_url}/api/v1/run/udf"

    # This is the body for when step_config_with_params.type == "udf".
    body = {
        "data_left": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    post_attr_json = body

    recursion_factor = get_recursion_factor()
    post_attr_headers = api._generate_headers({"Content-Type": "application/json"})
    post_attr_headers["Fused-Recursion"] = f"{recursion_factor}"

    req_params = {}

    # Make request
    start = time.time()

    if OPTIONS.pyodide_async_requests:
        import pyodide.http
        import yarl

        url_with_params = yarl.URL(url, encoded=True).with_query(req_params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            method="POST",
            headers=post_attr_headers,
            body=json.dumps(post_attr_json),
            # TODO: timeout
        )
    else:
        import aiohttp

        async with aiohttp.ClientSession() as session:
            r = await session.post(
                url=url,
                params=req_params,
                json=post_attr_json,
                headers=post_attr_headers,
                # TODO: timeout
            )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return await _process_response_async(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


def _process_response(
    r: requests.Response,
    step_config: AnyJobStepConfig,
    time_taken_seconds: float,
) -> UdfEvaluationResult:
    result_content: Optional[bytes] = None
    output_df: Optional[pd.DataFrame] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    error_message: Optional[str] = None
    error_lineno: Optional[int] = None

    try:
        if r.status_code != 200 and r.status_code != 422:
            raise ValueError(r.text)
        result_content = r.content
        # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
        _fused_metadata = r.headers.get("x-fused-metadata")
        fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

        # Extract stdout/stderr.
        stdout = fused_metadata.get("stdout")
        stderr = fused_metadata.get("stderr")

        # Extract udf.
        udf: Optional[AnyBaseUdf] = None
        if step_config is not None:
            udf = step_config.udf

        # Extract error line, if exists.
        error_lineno = fused_metadata.get("lineno")

        if r.status_code == 200:
            # If the UDF returned None.
            if len(result_content) == 0:
                return UdfEvaluationResult(
                    data=None,
                    udf=udf,
                    time_taken_seconds=time_taken_seconds,
                    stdout=stdout,
                    stderr=stderr,
                    error_message=error_message,
                    error_lineno=error_lineno,
                )

            # Else, process response output.
            res_buf = BytesIO(result_content)

            if r.headers["content-type"] == "image/png":
                import xarray as xr
                from PIL import Image

                display_method = infer_display_method(None, None)
                if display_method.show_widget:
                    from IPython.display import Image as IPythonImage
                    from IPython.display import display

                    display(IPythonImage(data=r.content, format="png"))

                image = Image.open(BytesIO(r.content))
                width, height = image.size
                if len(image.getbands()) == 1:
                    image_data = list(image.getdata())
                    image_data = [
                        image_data[i : i + width]
                        for i in range(0, len(image_data), width)
                    ]
                    data_array = xr.DataArray(image_data, dims=["y", "x"])
                else:
                    image_data = []
                    for band in range(len(image.getbands())):
                        band_data = list(image.getdata(band=band))
                        band_data = [
                            band_data[i : i + width]
                            for i in range(0, len(band_data), width)
                        ]
                        image_data.append(band_data)
                    data_array = xr.DataArray(image_data, dims=["band", "y", "x"])

                # Create the dataset with image, latitude, and longitude data
                dataset = xr.Dataset({"image": data_array})

                return UdfEvaluationResult(
                    data=dataset,
                    udf=udf,
                    time_taken_seconds=time_taken_seconds,
                    stdout=stdout,
                    stderr=stderr,
                    error_message=error_message,
                    error_lineno=error_lineno,
                )
            elif r.headers["content-type"] == "image/tiff":
                # TODO: Automatically display tiff?
                import rioxarray
                import xarray as xr

                with NamedTemporaryFile(prefix="udf_result", suffix=".tiff") as ntmp:
                    with open(ntmp.name, "wb") as f:
                        f.write(r.content)
                    rda = rioxarray.open_rasterio(
                        ntmp.name,
                        masked=True,
                    )
                    dataset = xr.Dataset({"image": rda})

                    return UdfEvaluationResult(
                        data=dataset,
                        udf=udf,
                        time_taken_seconds=time_taken_seconds,
                        stdout=stdout,
                        stderr=stderr,
                        error_message=error_message,
                        error_lineno=error_lineno,
                    )
            else:  # assume parquet
                m = pq.read_metadata(res_buf)
                if b"geo" in m.metadata:
                    try:
                        output_df = gpd.read_parquet(res_buf)
                    except ValueError as e:
                        warnings.warn(
                            f"Result has geo metadata but could not be loaded in GeoPandas: {e}",
                            FusedWarning,
                        )
                if output_df is None:
                    output_df = pd.read_parquet(res_buf)

                # if "fused_index" not in output_df.columns:
                #     # TODO: Hack since the backend no longer responds with fused_index
                #     output_df["fused_index"] = list(range(len(output_df)))

                new_output = _transform_output(
                    output=Output(data=output_df, skip_fused_index_validation=True)
                )

                new_output.validate_data_with_schema()

                if hasattr(udf, "table_schema") and udf.table_schema is None:
                    udf.table_schema = new_output.table_schema  # TODO: ?
                return UdfEvaluationResult(
                    data=new_output.data,
                    sidecar=new_output.sidecar_output,
                    udf=udf,
                    table_schema=new_output.table_schema,
                    time_taken_seconds=time_taken_seconds,
                    stdout=stdout,
                    stderr=stderr,
                    error_message=error_message,
                    error_lineno=error_lineno,
                )
        else:
            if "errormsg" in fused_metadata and fused_metadata["errormsg"]:
                error_message = f"The UDF returned the following error for chunk {fused_metadata['chunkinfo']} in line {fused_metadata.get('lineno')}:\n{fused_metadata['errormsg']}"
            elif "exception" in fused_metadata and fused_metadata["exception"]:
                error_message = fused_metadata["exception"]
            else:
                # No error message was returned, e.g. due to deserialization error
                try:
                    # Look for a "detail" field in the response payload
                    details_obj = json.loads(r.text)
                    error_message = details_obj["detail"]
                except:  # noqa: E722
                    error_message = r.text

            return UdfEvaluationResult(
                data=None,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=stdout,
                stderr=stderr,
                error_message=error_message,
                error_lineno=error_lineno,
            )
    except:  # noqa: E722
        r.raise_for_status()
        raise


async def _process_response_async(
    r,
    step_config: AnyJobStepConfig,
    time_taken_seconds: float,
) -> UdfEvaluationResult:
    result_content: Optional[bytes] = None
    output_df: Optional[pd.DataFrame] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    error_message: Optional[str] = None
    error_lineno: Optional[int] = None

    try:
        if r.status != 200 and r.status != 422:
            raise ValueError(await r.read())

        result_content = (
            await r.read() if not OPTIONS.pyodide_async_requests else await r.bytes()
        )
        # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
        _fused_metadata = r.headers.get("x-fused-metadata")
        fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

        # Extract stdout/stderr.
        stdout = fused_metadata.get("stdout")
        stderr = fused_metadata.get("stderr")

        # Extract udf.
        udf: Optional[AnyBaseUdf] = None
        if step_config is not None:
            udf = step_config.udf

        # Extract error line, if exists.
        error_lineno = fused_metadata.get("lineno")

        if r.status == 200:
            # If the UDF returned None.
            if len(result_content) == 0:
                return UdfEvaluationResult(
                    data=None,
                    udf=udf,
                    time_taken_seconds=time_taken_seconds,
                    stdout=stdout,
                    stderr=stderr,
                    error_message=error_message,
                    error_lineno=error_lineno,
                )

            # Else, process response output.
            res_buf = BytesIO(result_content)

            if r.headers["content-type"] == "image/png":
                import xarray as xr
                from PIL import Image

                display_method = infer_display_method(None, None)
                if display_method.show_widget:
                    from IPython.display import Image as IPythonImage
                    from IPython.display import display

                    display(IPythonImage(data=result_content, format="png"))

                image = Image.open(res_buf)
                width, height = image.size
                if len(image.getbands()) == 1:
                    image_data = list(image.getdata())
                    image_data = [
                        image_data[i : i + width]
                        for i in range(0, len(image_data), width)
                    ]
                    data_array = xr.DataArray(image_data, dims=["y", "x"])
                else:
                    image_data = []
                    for band in range(len(image.getbands())):
                        band_data = list(image.getdata(band=band))
                        band_data = [
                            band_data[i : i + width]
                            for i in range(0, len(band_data), width)
                        ]
                        image_data.append(band_data)
                    data_array = xr.DataArray(image_data, dims=["band", "y", "x"])

                # Create the dataset with image, latitude, and longitude data
                dataset = xr.Dataset({"image": data_array})

                return UdfEvaluationResult(
                    data=dataset,
                    udf=udf,
                    time_taken_seconds=time_taken_seconds,
                    stdout=stdout,
                    stderr=stderr,
                    error_message=error_message,
                    error_lineno=error_lineno,
                )
            elif r.headers["content-type"] == "image/tiff":
                # TODO: Automatically display tiff?
                import rioxarray
                import xarray as xr

                with NamedTemporaryFile(prefix="udf_result", suffix=".tiff") as ntmp:
                    with open(ntmp.name, "wb") as f:
                        f.write(result_content)
                    rda = rioxarray.open_rasterio(
                        ntmp.name,
                        masked=True,
                    )
                    dataset = xr.Dataset({"image": rda})

                    return UdfEvaluationResult(
                        data=dataset,
                        udf=udf,
                        time_taken_seconds=time_taken_seconds,
                        stdout=stdout,
                        stderr=stderr,
                        error_message=error_message,
                        error_lineno=error_lineno,
                    )
            else:  # assume parquet
                m = pq.read_metadata(res_buf)
                if b"geo" in m.metadata:
                    try:
                        output_df = gpd.read_parquet(res_buf)
                    except ValueError as e:
                        warnings.warn(
                            f"Result has geo metadata but could not be loaded in GeoPandas: {e}",
                            FusedWarning,
                        )
                if output_df is None:
                    output_df = pd.read_parquet(res_buf)

                # if "fused_index" not in output_df.columns:
                #     # TODO: Hack since the backend no longer responds with fused_index
                #     output_df["fused_index"] = list(range(len(output_df)))

                new_output = _transform_output(
                    output=Output(data=output_df, skip_fused_index_validation=True)
                )

                new_output.validate_data_with_schema()

                if hasattr(udf, "table_schema") and udf.table_schema is None:
                    udf.table_schema = new_output.table_schema  # TODO: ?
                return UdfEvaluationResult(
                    data=new_output.data,
                    sidecar=new_output.sidecar_output,
                    udf=udf,
                    table_schema=new_output.table_schema,
                    time_taken_seconds=time_taken_seconds,
                    stdout=stdout,
                    stderr=stderr,
                    error_message=error_message,
                    error_lineno=error_lineno,
                )
        else:
            if "errormsg" in fused_metadata and fused_metadata["errormsg"]:
                error_message = f"The UDF returned the following error for chunk {fused_metadata['chunkinfo']} in line {fused_metadata.get('lineno')}:\n{fused_metadata['errormsg']}"
            elif "exception" in fused_metadata and fused_metadata["exception"]:
                error_message = fused_metadata["exception"]
            else:
                # No error message was returned, e.g. due to deserialization error
                try:
                    # Look for a "detail" field in the response payload
                    details_obj = json.loads(result_content)
                    error_message = details_obj["detail"]
                except:  # noqa: E722
                    error_message = result_content.decode()

            return UdfEvaluationResult(
                data=None,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=stdout,
                stderr=stderr,
                error_message=error_message or "Unknown error occured",
                error_lineno=error_lineno,
            )
    except:  # noqa: E722
        r.raise_for_status()
        raise
